{-# OPTIONS --no-termination-check #-}

module Recursive where
open import Utils
open import Term
open import Value
open import Normal

infixl 50 _$$_

mutual
  ev : forall {Γ}{Δ}{σ} -> Tm Δ σ -> Env Γ Δ -> Val Γ σ
  ev ø            (_ ◂ v) = v
  ev (λ t)        vs      = λv t vs
  ev (t $ u)      vs      = ev t vs $$ ev u vs
  ev ( t [ ts ])   vs     = ev t (evˢ ts vs) 
  ev Z            _       = Zv
  ev (S n)        vs      = Sv (ev n vs)
  ev (pr n f z)   vs      = primv (ev n vs) (ev f vs) (ev z vs)

  primv : forall {Γ}{σ} -> Val Γ N -> Val Γ (N → σ → σ) -> Val Γ σ -> Val Γ σ
  primv Zv      _ z = z 
  primv (Sv n)  f z = f $$ n $$ primv n f z
  primv (nev n) f z = nev (npr n f z)  

  _$$_ : forall {Γ}{σ}{τ} -> Val Γ (σ → τ) -> Val Γ σ -> Val Γ τ
  λv t vs $$ v = ev t (vs ◂ v)
  nev n   $$ v = nev (app n v) 

  evˢ : forall {B}{Γ}{Δ} -> Tmˢ Γ Δ -> Env B Γ -> Env B Δ
  evˢ id        vs       = vs
  evˢ (ts ◃ t)  vs       = evˢ ts vs ◂ ev t vs
  evˢ (↑ _)     (vs ◂ _) = vs
  evˢ (ts ○ us) vs       = evˢ ts (evˢ us vs)

mutual
  quote : forall {Γ}{σ} -> Val Γ σ -> Nf Γ σ
  quote {Γ}{σ → τ} f       = λn (quote (wkv _ f $$ (nev (var vZ)))) 
  quote {Γ}{N}     (nev n) = nen (quoteⁿ n)
  quote {Γ}{N}     Zv      = Zn 
  quote {Γ}{N}     (Sv n)  = Sn (quote n) 

  quoteⁿ : forall {Γ}{σ} -> Ne Val Γ σ -> Ne Nf Γ σ
  quoteⁿ (var v)     = var v
  quoteⁿ (app n v)   = app (quoteⁿ n) (quote v)
  quoteⁿ (npr n f z) = npr (quoteⁿ n) (quote f) (quote z)

nf : forall {Γ}{σ} -> Tm Γ σ -> Nf Γ σ
nf t = quote (ev t ide)